/*
   Class Name:

      CDirectInputSystem.

   Created by:

      Allen Sherrod (Programming Ace of www.UltimateGameProgramming.com).

   Description:

      This class is used with DirectInput to get input from a keyboard and/or mouse.
*/


#include"CDirectInput.h"      // Our input system, keyboard, and mouse class.


CDirectInputSystem::CDirectInputSystem()
{
   // Initialize objects...
   m_Keyboard = NULL;
   m_Mouse = NULL;
   memset(&m_Mouse_State, 0, sizeof(m_Mouse_State));
   memset(&m_oldMouse_State, 0, sizeof(m_oldMouse_State));
   memset(m_Keys, 0, sizeof(m_Keys));
   memset(m_oldKeys, 0, sizeof(m_oldKeys));
}


CDirectInputSystem::~CDirectInputSystem()
{
   // Shut everything down.
   Shutdown();
}


bool CDirectInputSystem::Initialize(HWND hwnd, HINSTANCE hInstance, bool mouseExclusive)
{
   // Now we are ready to create the input system.  Call DirectInput8Create() function
   // to create the Direct Input object.  This function needs the HINSTANCE object, the
   // version, what is called a unique identifier, the LPDIRECTINPUT8 object, and
   // the last field is used for COM.  Check out the DirectX SDK for more on that one.
   if(FAILED(DirectInput8Create(hInstance, DIRECTINPUT_VERSION, IID_IDirectInput8,
                                (void **)&m_InputSystem, NULL)))
      return false;

   // Initialize the keyboard.

   // First lets create the device.  CreateDevice() requires a ID, input device object,
   // and a little something for COM.  The last can be NULL.
   if(FAILED(m_InputSystem->CreateDevice(GUID_SysKeyboard, &m_Keyboard, NULL)))
      return false;

   // Here we need to Set the data format.  This will tell direct input what kind of device
   // this is.  This must be called before we can gain access to the device.
   if(FAILED(m_Keyboard->SetDataFormat(&c_dfDIKeyboard))) return false;

   // Now we will set the flags and options for our keyboard object.
   if(FAILED(m_Keyboard->SetCooperativeLevel(hwnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE)))
      return false;

   // Now we call Acquire() to gain access to our input device.
   if(FAILED(m_Keyboard->Acquire())) return false;

   // ClearKeys() will clear out the array of keys we have.
   ZeroMemory(m_Keys, KEYS_SIZE * sizeof(char));


   // Initialize the Mouse.
   DWORD flags;

   // Create device, set data format, and set flags like we did with the keyboard.
   if(FAILED(m_InputSystem->CreateDevice(GUID_SysMouse, &m_Mouse, NULL))) return false;
   if(FAILED(m_Mouse->SetDataFormat(&c_dfDIMouse))) return false;

   if(mouseExclusive) flags = DISCL_FOREGROUND | DISCL_EXCLUSIVE | DISCL_NOWINKEY;
   else flags = DISCL_FOREGROUND | DISCL_NONEXCLUSIVE;

   if(FAILED(m_Mouse->SetCooperativeLevel(hwnd, flags))) return false;

   // Gain access.
   if(FAILED(m_Mouse->Acquire())) return false;

   return true;
}


bool CDirectInputSystem::UpdateDevices()
{
   if(!m_Mouse || !m_Keyboard) return false;

   memcpy(&m_oldMouse_State, &m_Mouse_State, sizeof(m_Mouse_State));

   // Get the device state.  If there was an error re-acquire the device and try again.
   if(FAILED(m_Mouse->GetDeviceState(sizeof(DIMOUSESTATE), &m_Mouse_State)))
      {
         if(FAILED(m_Mouse->Acquire())) return false;
         if(FAILED(m_Mouse->GetDeviceState(sizeof(DIMOUSESTATE), &m_Mouse_State)))
            return false;
      }

   // Copy old keys so we can determine when a button is raised.
   memcpy(m_oldKeys, m_Keys, sizeof(m_Keys));

   if(FAILED(m_Keyboard->GetDeviceState(sizeof(m_Keys), (LPVOID)m_Keys)))
      {
         if(FAILED(m_Keyboard->Acquire())) return false;
         if(FAILED(m_Keyboard->GetDeviceState(sizeof(m_Keys), (LPVOID)m_Keys)))
            return false;
      }

   return true;
}


bool CDirectInputSystem::KeyUp(int key)
{
   // If the key in the variable key is not pressed then we would return false, else trye.
   return (!(m_Keys[key] & 0x80) && m_Keys[key] != m_oldKeys[key]);
}


bool CDirectInputSystem::KeyDown(int key)
{
   // If the key in the variable key is not pressed then we would return true, else false.
   return m_Keys[key] & 0x80;
}


bool CDirectInputSystem::MouseButtonUp(int button)
{
   // If the button in question is not clicked we return false, else true.
   return (!(m_Mouse_State.rgbButtons[button] & 0x80) &&
           m_Mouse_State.rgbButtons[button] != m_oldMouse_State.rgbButtons[button]);
}


bool CDirectInputSystem::MouseButtonDown(int button)
{
   // If the button in question is clicked we return true, else false.
   return m_Mouse_State.rgbButtons[button] & 0x80;
}


void CDirectInputSystem::Shutdown()
{
   // Delete each object...
   if(m_Keyboard)
      {
         m_Keyboard->Unacquire();
         m_Keyboard->Release();
         m_Keyboard = NULL;
      }

   if(m_Mouse)
      {
         m_Mouse->Unacquire();
         m_Mouse->Release();
         m_Mouse = NULL;
      }

   // Release all of the resources out Direct Input object used.
   if(m_InputSystem)
      {
         m_InputSystem->Release();
         m_InputSystem = NULL;
      }
}


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com