/*
   Direct Input Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to process input using Direct Input (which
   is apart of DirectX).  This method is much faster because you will have access
   to the hardware Directly plus a lot of the behind the scenes work is hidden so
   dealing with input is much easier.  What this demo will do is whenever the
   user presses a mouse button the object being rendered will change.  Also the
   user can press the 1, 2, 3, or 4 key to change the objects with the keyboard.

   To get access to the keyboard and mouse we must create two objects of type
   LPDIRECTINPUTDEVICE8.  To gain access to Direct Input we need an object of type
   LPDIRECTINPUT8.  Once we create the Direct Input object using the function
   DirectInput8Create() you can then create each device using the function
   CreateDevice().  After that the rest of the setup involves setting flags and options
   using the SetCooperativeLevel() function and to gain access to the device using the
   Acquire() function.  Thats it.  When your program is executing you will update
   that states of each device to test if a button was pressed or if the joystick or mouse
   has moved, etc.  When it is all done you call the Shutdown() function to delete Direct
   Input and all its devices.  All of the code for it is in the CDirectInput.cpp file.
*/


// Demo: Direct Input.


#define WIN32_LEAN_AND_MEAN// Trims down the libraries.
#define VC_LEANMEAN        // Trims even further.


// Defines for which object we will display.  We will use our new input system to cycle
// through the list.
#define CUBE   0
#define SPHERE 1
#define TORUS  2
#define TEAPOT 3


#include<windows.h>           // Windows header file.
#include<gl/gl.h>             // OpenGL header.
#include<gl/glu.h>            // OpenGL utility.
#include<gl/glaux.h>          // Yet another OpenGL header.
#include"CDirectInput.h"      // Input class.


// Function prototypes...
bool InitializeGL();          // Initialize OpenGL.
void RenderScene();           // Renders each frame.
void GetInput();              // Check the input for each frame.


// The first is a global HDC that will be used in the RenderScene() function to swap buffers.
HDC g_HDC;


// Light properties for the diffuse light, specular light, and light position.
float diffuseLight[] = {0.5, 0.5, 0.5, 1.0};
float specularLight[] = {1.0, 1.0, 1.0, 1.0};
float LightPosition[] = {0.0, 0.0, 10.0, 1.0};

// These are the x and y rotations for our objects.
float xRotation = 0.0f;
float yRotation = 0.0f;

// This will allow us to cycle through different objects for display.
int ObjectChoice = 0;

// Here is our Direct Input system object used to check for user input.
CDirectInputSystem DirectInput;


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // this chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;
   static HDC hDC;
   int width, height; 


   switch(message)
      {
         case WM_CREATE:                     // Windows creation.
            hDC = GetDC(hwnd);               // This gets the device context for our window.
            g_HDC = hDC;                     // Assigns the global device context to this one.
            SetupPixelFormat(hDC);           // Call the pixel format function.

            hRC = wglCreateContext(hDC);     // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);        // Makes the rendering context.
            return 0;
            break;

         case WM_CLOSE:                      // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);           // Deletes the rendering context.

            PostQuitMessage(0);              // Says close the program.
            return 0;
            break;

         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height); // resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);     // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            // calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);      // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            return 0;
            break;

         default: // Always have a default just in case.
            break;
      }

   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;
   WNDCLASSEX windowClass;
   HWND hwnd;
   bool isFinished;

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create window.
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "Direct Input Demo - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |// window style.
                        WS_CLIPSIBLINGS,// window style.
                        100, 100,// window x, y coordinate.
                        640,
                        480,// window width and height.
                        NULL,// handle to parent window.
                        NULL,// menu.
                        hInstance,// handle to app instance.
                        NULL);   // pointer to window creation data.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.


   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL()) isFinished = true;

   // Here is where we set up Direct Input so we can get user input from
   // the mouse & Keyboard.
   if(!DirectInput.Initialize(hwnd, hInstance)) return false;

   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);  // Translate any messages.
               DispatchMessage(&msg);   // Dispatch any messages.
            }
         else
            {
               RenderScene();          // Render a frame.
            }
      }

   // Shutdown Direct Input.
   DirectInput.Shutdown();

   return (int)msg.wParam;      // End the application.
}


bool InitializeGL()
{
   glClearColor(0.0f, 0.0f, 0.0f, 0.0f);  // Clear the screen to black.

   glShadeModel(GL_SMOOTH);               // Smooth shading in our scenes.
   glEnable(GL_DEPTH_TEST);               // Enable desth testing for hidden surface removal.
   glEnable(GL_LIGHTING);                 // Enable lighting.


   // Here we will set up a spotlight light source in our scene.
   glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
   glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight);
   glLightfv(GL_LIGHT0, GL_POSITION, LightPosition);
   glLightf(GL_LIGHT0, GL_SPOT_CUTOFF, 10.0f);
   glLightf(GL_LIGHT0, GL_SPOT_EXPONENT, 75.0f);

   // Enable our light.
   glEnable(GL_LIGHT0);

   // Set up the material information for our objects.  Again this is just for show.
   glEnable(GL_COLOR_MATERIAL);
   glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
   glMaterialfv(GL_FRONT, GL_SPECULAR, specularLight);
   glMateriali(GL_FRONT, GL_SHININESS, 128);

   // If all went well we return true.
   return true;
}


void RenderScene()
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // Clears the screen.
   glLoadIdentity();                   //Reset modelview matrix for new frame.

   // First lets test our input objects and see if the user has pressed any key we care
   // about or has click any of the mouse buttons.
   GetInput();

   glTranslatef(0.0, 0.0, -5.0f);            // Move back 5 units.

   glRotatef(yRotation, 0.0f, 1.0f, 0.0f);   // Rotate the object along the y axis.
   glRotatef(xRotation, 1.0f, 0.0f, 0.0f);   // Rotate the object along the x axis.

   // Draw our object.  Depending on what ObjectChoice is will depend on what is drawn...
   switch(ObjectChoice)
      {
         case CUBE:
            auxSolidCube(2.0);               // Draw a solid cube.
            break;

         case SPHERE:
            auxSolidSphere(1.5);             // Draw a solid sphere.
            break;

         case TORUS:
            auxSolidTorus(0.5, 1.0);         // Draw a solid torus.
            break;

         case TEAPOT:                        // Draw a wireframe teapot.
            auxWireTeapot(1.5);
            break;

         default:                            // Reset in case neither is true.
            ObjectChoice = 0;
            break;
      }

   // Increase the rotation.
   xRotation += 0.1f;
   yRotation += 0.2f;

   // Error checking to make sure our values do not go over 360 degrees.
   if(xRotation >= 360.0f) xRotation = 0.0f;
   if(yRotation >= 360.0f) yRotation = 0.0f;

   SwapBuffers(g_HDC);
}


void GetInput()
{
   // First we will update all the devices our input system supports.  This will
   // check which keys are pressed and which mouse buttons are pressed.

   // Update all devices.
   DirectInput.UpdateDevices();

   // Next we will check for the keys and buttons we care about.

   // Check if the user pressed the escape key.
   if(DirectInput.KeyUp(DIK_ESCAPE)) PostQuitMessage(0);

   // Check if the user pressed the number 1 key.
   if(DirectInput.KeyUp(DIK_1)) ObjectChoice = CUBE;

   // Check if the user pressed the number 2 key.
   if(DirectInput.KeyUp(DIK_2)) ObjectChoice = SPHERE;

   // Check if the user pressed the number 3 key.
   if(DirectInput.KeyUp(DIK_3)) ObjectChoice = TORUS;

   // Check if the user pressed the number 4 key.
   if(DirectInput.KeyUp(DIK_4)) ObjectChoice = TEAPOT;

   // Check if the user pressed the left mouse button.  If so increase the ObjectChoice variable.
   if(DirectInput.MouseButtonUp(LEFT_BUTTON)) ObjectChoice++;

   // Check if the user pressed the right mouse button.  If so decrease the ObjectChoice variable.
   if(DirectInput.MouseButtonUp(RIGHT_BUTTON)) ObjectChoice--;

   // Error checking.
   if(ObjectChoice >= 4) ObjectChoice = CUBE;
   if(ObjectChoice < 0) ObjectChoice = TEAPOT;
}


// Recap:

// Well there you have it.  Getting input through the keyboard and mouse using Direct Input.
// As you can see it is not that hard.  This is the better way to get input because using
// Direct Input you are accessing the hardware directly.  Another good thing about Direct Input
// is that you can easily add more devices, such as a joystick, without any complex code.


// Copyright October 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com